require 'stbaldricks/endpoints/donation'
require 'stbaldricks/entities/fund'
require 'stbaldricks/entities/event'
require 'stbaldricks/entities/team'
require 'stbaldricks/entities/participant'
require 'stbaldricks/entities/fundraiser'
require 'stbaldricks/entities/person'
require 'stbaldricks/entities/organization'
require 'stbaldricks/entities/memorial'
require 'stbaldricks/entities/lib/address'
require 'stbaldricks/entities/lib/top_level'
require 'stbaldricks/entities/lib/name'
require 'stbaldricks/entities/lib/payment'
require 'stbaldricks/entities/lib/default_cacheable'
require 'stbaldricks/entities/concerns/type_concern'
require 'stbaldricks/entities/concerns/model_type_concern'

module SBF
  module Client
    class Donation < SBF::Client::TopLevelEntity
      include Entities::DefaultCacheable
      endpoint SBF::Client::DonationEndpoint
      actions DEFAULT_CRUD_ACTIONS
      action :external_sync

      disallow_instantiation
      class EmployerMatching < SBF::Client::BaseEntity
        attr_accessor :is_match, :employer, :matched_donation_id

        def matched_donation_id=(value)
          matched_donation_id_will_change!
          @matched_donation_id = value.to_i
        end
      end

      class EmployerMatched < SBF::Client::BaseEntity
        attr_accessor :is_match
        entity_attr_accessor :matching_donation, 'SBF::Client::FullDonation'
      end

      class Tribute < SBF::Client::BaseEntity
        attr_accessor :type
        attr_reader :name
        entity_attr_accessor :name_pieces, 'SBF::Client::NamePieces'

        module Type
          HONOR = 'honor'.freeze
          TRIBUTE = 'tribute'.freeze
          MEMORY = 'memory'.freeze
        end
      end

      class FullParticipant < SBF::Client::FullParticipant
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::PARTICIPANT
        end
      end

      class PartialParticipant < SBF::Client::PartialParticipant
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::PARTICIPANT
        end
      end

      class FullFundraiser < SBF::Client::FullFundraiser
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::FUNDRAISER
        end
      end

      class PartialFundraiser < SBF::Client::PartialFundraiser
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::FUNDRAISER
        end
      end

      class FullTeam < SBF::Client::FullTeam
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::TEAM
        end
      end

      class PartialTeam < SBF::Client::PartialTeam
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::TEAM
        end
      end

      class FullEvent < SBF::Client::FullEvent
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::EVENT
        end
      end

      class PartialEvent < SBF::Client::PartialEvent
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::EVENT
        end
      end

      class FullFund < SBF::Client::FullFund
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::FUND
        end
      end

      class PartialFund < SBF::Client::PartialFund
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::FUND
        end
      end

      class FullMemorial < SBF::Client::FullMemorial
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::MEMORIAL
        end
      end

      class PartialMemorial < SBF::Client::PartialMemorial
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::Donation::RecipientType::MEMORIAL
        end
      end

      class SBFGeneralFund < SBF::Client::BaseEntity
        include ModelTypeConcern
        attr_reader :model_type

        def id
          nil
        end

        def model_type
          SBF::Client::Donation::RecipientType::SBF_GENERAL_FUND
        end
      end

      class FullPerson < SBF::Client::FullPerson
        include TypeConcern
        attr_reader :type

        def type
          SBF::Client::Donation::DonorType::INDIVIDUAL
        end
      end

      class PartialPerson < SBF::Client::PartialPerson
        include TypeConcern
        attr_reader :type

        def type
          SBF::Client::Donation::DonorType::INDIVIDUAL
        end
      end

      class FullOrganization < SBF::Client::FullOrganization
        include TypeConcern
        attr_reader :type
      end

      class PartialOrganization < SBF::Client::PartialOrganization
        include TypeConcern
        attr_reader :type
      end

      class RecognitionCard < SBF::Client::BaseEntity
        class Recipient < SBF::Client::BaseEntity
          attr_reader :name
          entity_attr_accessor :name_pieces, SBF::Client::NamePieces
          entity_attr_accessor :address, SBF::Client::Address
        end

        entity_attr_accessor :recipient, SBF::Client::Donation::RecognitionCard::Recipient
        attr_accessor :salutation, :message, :signature
      end

      def self.generate_special_donation_type(entity)
        return if entity.nil?

        # Calculate the country and state
        country, state = get_country_and_state_for_special_donation_type(entity)

        # state only applicable for United States, others only use international value
        unless country.empty?
          if country.to_sym == SBF::Client::Country.list[SBF::Client::Country::UNITED_STATES]
            state_value = state.to_s.upcase.to_sym
            if !state_value.empty? && SBF::Client::Donation::SpecialDonationType.const_defined?(state_value)
              return SBF::Client::Donation::SpecialDonationType.const_get(state_value)
            end

            return SBF::Client::Donation::SpecialDonationType::NO_STATE
          else
            international_value = "INT_#{country}".upcase.to_sym
            if SBF::Client::Donation::SpecialDonationType.const_defined?(international_value)
              return SBF::Client::Donation::SpecialDonationType.const_get(international_value)
            end

            return SBF::Client::Donation::SpecialDonationType::INTERNATIONAL
          end
        end

        SBF::Client::Donation::SpecialDonationType::GENERAL_EVENT
      end

      def self.get_country_and_state_for_special_donation_type(entity)
        case entity
        when SBF::Client::Search::Participant, SBF::Client::Search::Fundraiser, SBF::Client::Search::Team, SBF::Client::Search::Event
          if entity.event_id == 1
            # Look up the person since it is a Search::Participant and doesn't have sub-objects
            response = SBF::Client::Person.get(entity.profile_id)
            person = response.data
            raise SBF::Client::Error, 'Invalid Entity' if response.error? || person.nil?

            # Get country and state from the person
            country = person.addresses.primary.country
            state = (country.to_sym == SBF::Client::Country.list[SBF::Client::Country::UNITED_STATES] ? person.addresses.primary.state : '')

          else
            country = (defined?(entity.country).nil? ? nil : entity.country)
            state = (defined?(entity.state).nil? ? nil : entity.state)

          end

          [country, state]

        when SBF::Client::Participant
          get_event_country_and_state_for_special_donation_type(entity.event, entity.person)

        when SBF::Client::Fundraiser
          [entity.venue.location.address.country, entity.venue.location.address.state]

        when SBF::Client::Team
          get_event_country_and_state_for_special_donation_type(entity.event)

        when SBF::Client::Event
          get_event_country_and_state_for_special_donation_type(entity)

        end
      end
      private_class_method :get_country_and_state_for_special_donation_type

      def self.get_event_country_and_state_for_special_donation_type(event, person = nil)
        if event.virtual?
          raise SBF::Client::Error, 'Invalid Entity' if person.nil?

          # Get country and state from the person
          country = person.addresses.primary.country
          state = (country.to_sym == SBF::Client::Country.list[SBF::Client::Country::UNITED_STATES] ? person.addresses.primary.state : '')

        else
          country = (defined?(event.venue.location.address.country).nil? ? nil : event.venue.location.address.country)
          state = (defined?(event.venue.location.address.state).nil? ? nil : event.venue.location.address.state)

        end

        [country, state]
      end
      private_class_method :get_event_country_and_state_for_special_donation_type
    end

    class PartialDonation < SBF::Client::Donation
      attr_accessor :id
    end
    class FullDonation < SBF::Client::Donation
      attr_accessor :id
      attr_accessor :status
      attr_accessor :amount
      attr_accessor :fee_amount
      attr_accessor :display_name
      attr_accessor :year
      attr_reader :is_send_card

      multitype_attr_accessor(
        :recipient,
        [
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::PARTICIPANT },
            'SBF::Client::Donation::FullParticipant',
            'SBF::Client::Donation::PartialParticipant'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::FUNDRAISER },
            'SBF::Client::Donation::FullFundraiser',
            'SBF::Client::Donation::PartialFundraiser'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::TEAM },
            'SBF::Client::Donation::FullTeam',
            'SBF::Client::Donation::PartialTeam'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::EVENT },
            'SBF::Client::Donation::FullEvent',
            'SBF::Client::Donation::PartialEvent'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::FUND },
            'SBF::Client::Donation::FullFund',
            'SBF::Client::Donation::PartialFund'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::MEMORIAL },
            'SBF::Client::Donation::FullMemorial',
            'SBF::Client::Donation::PartialMemorial'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::Donation::RecipientType::SBF_GENERAL_FUND },
            'SBF::Client::Donation::SBFGeneralFund'
          ]
        ],
        optional: true
      )

      attr_accessor :is_unrecognized, :special_donation_type
      multitype_attr_accessor(
        :donor,
        [
          [
            ->(v) { v[:type] == SBF::Client::Donation::DonorType::INDIVIDUAL },
            'SBF::Client::Donation::FullPerson',
            'SBF::Client::Donation::PartialPerson'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Donation::DonorType::BUSINESS },
            'SBF::Client::Donation::FullOrganization',
            'SBF::Client::Donation::PartialOrganization'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Donation::DonorType::ORGANIZATION_FOUNDATION },
            'SBF::Client::Donation::FullOrganization',
            'SBF::Client::Donation::PartialOrganization'
          ]
        ],
        optional: true
      )

      multitype_attr_accessor(
        :submitter,
        [
          [
            ->(v) { v[:type] == SBF::Client::Donation::DonorType::INDIVIDUAL },
            'SBF::Client::FullPerson',
            'SBF::Client::PartialPerson'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Donation::DonorType::BUSINESS },
            'SBF::Client::FullOrganization',
            'SBF::Client::PartialOrganization'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Donation::DonorType::ORGANIZATION_FOUNDATION },
            'SBF::Client::FullOrganization',
            'SBF::Client::PartialOrganization'
          ]
        ],
        optional: true
      )

      entity_attr_accessor :employer_matching, 'SBF::Client::Donation::EmployerMatching'
      entity_attr_accessor :employer_matched, 'SBF::Client::Donation::EmployerMatched'
      attr_accessor :foreign_beneficiary_country
      attr_accessor :promotional_code
      entity_attr_accessor :recurring_gift, 'SBF::Client::FullRecurringGift'
      entity_attr_accessor :fund, 'SBF::Client::FullFund'
      entity_attr_accessor :memorial, 'SBF::Client::FullMemorial'
      multitype_attr_accessor(
        :payment_details,
        [
          [->(v) { v[:type] == SBF::Client::Payment::Type::APPLEPAY }, 'SBF::Client::Payment::ApplepayDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::CREDIT_CARD }, 'SBF::Client::Payment::CreditCardDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::CASH }, 'SBF::Client::Payment::CashDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::CHECK }, 'SBF::Client::Payment::CheckDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::FACEBOOK }, 'SBF::Client::Payment::FacebookDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::GOOGLEPAY }, 'SBF::Client::Payment::GooglePayDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::MONEY_ORDER }, 'SBF::Client::Payment::MoneyOrderDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::NONCE }, 'SBF::Client::Payment::NonceDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::PAYPAL }, 'SBF::Client::Payment::PaypalDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::STRIPE }, 'SBF::Client::Payment::StripeDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::UNKNOWN }, 'SBF::Client::Payment::UnknownDetails'],
          [->(v) { v[:type] == SBF::Client::Payment::Type::WIRE_TRANSFER }, 'SBF::Client::Payment::WireTransferDetails']
        ]
      )
      entity_attr_accessor :tribute, 'SBF::Client::Donation::Tribute'
      entity_attr_accessor :recognition_card, 'SBF::Client::Donation::RecognitionCard'
      attr_accessor :collected_at
      attr_accessor :created_at, :modified_at, :modified_by, :how_created
      attr_accessor :is_thanked

      def amount=(value)
        amount_will_change!
        @amount = value.to_s.gsub(/[^0-9\.]/, '').to_f
      end

      def fee_amount=(value)
        fee_amount_will_change!
        @fee_amount = value.to_s.gsub(/[^0-9\.]/, '').to_f
      end
    end
  end
end
