require 'stbaldricks/endpoints/event'
require 'stbaldricks/entities/fund'
require 'stbaldricks/entities/campaign'
require 'stbaldricks/entities/person'
require 'stbaldricks/entities/participant'
require 'stbaldricks/entities/shave_schedule'
require 'stbaldricks/entities/lib/top_level'
require 'stbaldricks/entities/lib/cacheable'
require 'stbaldricks/entities/lib/email_address'
require 'stbaldricks/entities/lib/fundraising_page'
require 'stbaldricks/entities/lib/name'
require 'stbaldricks/entities/lib/phone'
require 'stbaldricks/entities/lib/venue'
require 'stbaldricks/entities/lib/geo_location'
require 'stbaldricks/entities/lib/permissions'
require 'stbaldricks/entities/lib/third_party_media'
require 'stbaldricks/entities/concerns/donation_recipient_concern'
require 'stbaldricks/entities/concerns/event_year_concern'
require 'stbaldricks/entities/concerns/fundraising_page_concern'
require 'stbaldricks/entities/concerns/venue_concern'

module SBF
  module Client
    class Event < SBF::Client::TopLevelEntity
      include VenueConcern, FundraisingPageConcern, EventYearConcern, DonationRecipientConcern, Entities::Cacheable
      endpoint SBF::Client::EventEndpoint
      actions DEFAULT_CRUD_ACTIONS
      action :invite

      cache_id -> { [id, year] }
      cache_id_from_hash ->(h) { [h[:event_id], h[:event_year]] }

      disallow_instantiation

      def update(with = {})
        endpoint.update(id, year, self, with)
      end

      def keys_hash
        year.nil? ? super : super.merge(year: year)
      end

      class Photos < SBF::Client::BaseEntity
        attr_reader :avatar, :default, :icon
      end

      class Contacts < SBF::Client::BaseEntity
        class Contact < SBF::Client::BaseEntity
          attr_reader :name
          entity_attr_accessor :name_pieces, 'SBF::Client::Event::Contacts::NamePieces'
          entity_attr_accessor :phone, 'SBF::Client::Phone'
          entity_attr_accessor :email_address, 'SBF::Client::EmailAddress'
        end

        entity_attr_accessor :primary, 'SBF::Client::Event::Contacts::Contact'
        entity_attr_accessor :secondary, 'SBF::Client::Event::Contacts::Contact'

        class NamePieces < NamePiecesBase
          attr_accessor :prefix, :first_name, :last_name, :suffix
        end
      end

      class Activity < SBF::Client::BaseEntity
        attr_reader :id
        entity_attr_accessor :event, 'SBF::Client::FullEvent', 'SBF::Client::PartialEvent', true
        attr_accessor :start_time
        attr_accessor :end_time
        attr_accessor :activity
        attr_accessor :sort_order
      end

      class CoachTracking < SBF::Client::BaseEntity
        class Proceeds < SBF::Client::BaseEntity
          attr_accessor :received_date, :notification_email_sent_date
        end

        class CoachingInteractions < SBF::Client::BaseEntity
          attr_accessor :introductory_conversation_date, :post_event_conversation_date
        end

        class Plaque < SBF::Client::BaseEntity
          attr_accessor :issued_to, :sent_date
        end

        attr_accessor :host_agreement_date, :activated_date, :proceeds, :coaching_interactions, :plaque
        entity_attr_accessor :proceeds, 'SBF::Client::Event::CoachTracking::Proceeds'
        entity_attr_accessor :coaching_interactions, 'SBF::Client::Event::CoachTracking::CoachingInteractions'
        entity_attr_accessor :plaque, 'SBF::Client::Event::CoachTracking::Plaque'
      end

      class Totals < SBF::Client::BaseEntity
        attr_reader :fundraisers, :participants, :shavees, :amount_raised
        attr_reader :current_year_ytd, :previous_year_ytd, :previous_year_total
        attr_reader :current_year_shavees_ytd, :previous_year_shavees_ytd, :previous_year_shavees_total
      end

      class Agreement < SBF::Client::BaseEntity
        module Checkbox
          CHECKED = 1
        end

        module AmentityAnswerOption
          NONE = 0
          PROVIDED_BY_VENUE = 1
          ARRANGED_FOR_BY_ORGANIZER_OR_TEAM = 2
          DONATED = 3
        end

        class Question < SBF::Client::BaseEntity
          attr_accessor :question_id, :question, :answer
        end

        attr_accessor :agreement_received_date
        entity_collection_attr_accessor :questions, 'SBF::Client::Event::Agreement::Question'
      end

      class Venue < SBF::Client::Venue
        class Social < SBF::Client::BaseEntity
          attr_accessor :facebook_url, :twitter_name
        end

        class Location < SBF::Client::Location
          module Type
            RESTAURANT = 'restaurant'.freeze
            SCHOOL = 'school'.freeze
            BUSINESS = 'business'.freeze
            PRIVATE_RESIDENCE = 'private_residence'.freeze
            HOSPITAL = 'hospital'.freeze
            MUNICIPALITY = 'municipality'.freeze
            OTHER = 'other'.freeze
            LEGACY = 'legacy'.freeze
            VIRTUAL = 'virtual'.freeze
            HYBRID = 'hybrid'.freeze
          end

          attr_accessor :type
          attr_accessor :type_other
        end

        entity_attr_accessor :location, 'SBF::Client::Event::Venue::Location'
        entity_attr_accessor :phone, 'SBF::Client::Phone'
        attr_accessor :website_url
        entity_attr_accessor :social, 'SBF::Client::Event::Venue::Social'
      end
    end

    class PartialEvent < Event
      attr_accessor :id, :year

      # TODO: Having this check be different between partial/full is bad and should be fixed.
      def virtual?
        (id == 1)
      end
    end

    class FullEvent < Event
      extend Forwardable
      def_delegator :@photos, :avatar, :avatar_photo_url

      attr_accessor :id, :year
      attr_accessor :status
      attr_reader :display_name
      attr_accessor :event_title
      attr_accessor :type
      attr_accessor :type_other
      attr_accessor :sub_type
      attr_accessor :is_private
      attr_accessor :date, :is_date_confirmed
      attr_accessor :start_time, :end_time, :is_time_confirmed
      attr_accessor :is_open_for_registration, :is_open_for_fast_registration
      attr_accessor :offline_donation_uuid, :offline_donation_form
      attr_accessor :fundraising_goal, :fundraising_message
      entity_attr_accessor :geo_location, 'SBF::Client::GeoLocation'
      entity_attr_accessor :fundraising_page, 'SBF::Client::FundraisingPage', nil, true
      entity_attr_accessor :campaign, 'SBF::Client::FullCampaign', 'SBF::Client::PartialCampaign', true
      entity_attr_accessor :fund, 'SBF::Client::FullFund', 'SBF::Client::PartialFund', true
      entity_attr_accessor :venue, 'SBF::Client::Event::Venue'
      entity_attr_accessor :photos, 'SBF::Client::Event::Photos'
      entity_attr_accessor :contacts, 'SBF::Client::Event::Contacts', nil, true
      entity_attr_accessor :coach_tracking, 'SBF::Client::Event::CoachTracking'
      entity_attr_accessor :shave_schedule, 'SBF::Client::ShaveSchedule', nil, true
      entity_attr_reader :totals, 'SBF::Client::Event::Totals', nil, true
      entity_collection_attr_accessor :coaches, 'SBF::Client::FullPerson', nil, true
      entity_collection_attr_accessor :closed_roles, 'SBF::Client::Participant::Roles::Type', nil, true
      attr_reader :created_at, :modified_at
      attr_accessor :modified_by
      attr_accessor :how_created
      attr_accessor :promotional_code
      entity_attr_reader :permissions, 'SBF::Client::Permissions', nil, true
      entity_attr_reader :agreement, 'SBF::Client::Event::Agreement', nil, true
      entity_collection_attr_accessor :organizers, 'SBF::Client::FullParticipant', nil, true
      entity_collection_attr_accessor :treasurers, 'SBF::Client::FullParticipant', nil, true
      entity_collection_attr_accessor :activities, 'SBF::Client::Event::Activity', nil, true
      entity_attr_accessor :third_party_media, 'SBF::Client::ThirdPartyMedia', nil, true
      entity_collection_attr_accessor :supporters, 'SBF::Client::FullEventSupporter', nil, true
      attr_accessor :fund_relationship_type

      def active?
        status == SBF::Client::Event::Status::ACTIVE
      end

      def venue_virtual?
        [SBF::Client::Event::Venue::Location::Type::VIRTUAL,
         SBF::Client::Event::Venue::Location::Type::HYBRID].include? venue.location.type
      end

      def virtual?
        (id == 1) && (sub_type == SBF::Client::Event::SubType::VIRTUAL)
      end

      def open_for_registration?
        is_open_for_registration
      end

      def open_for_fast_registration?
        is_open_for_fast_registration
      end

      def past?
        Date.today > Date.parse(date)
      rescue
        nil
      end
    end
  end
end
