require 'stbaldricks/entities/lib/top_level'
require 'stbaldricks/endpoints/fundraiser'
require 'stbaldricks/entities/person'
require 'stbaldricks/entities/fund'
require 'stbaldricks/entities/event'
require 'stbaldricks/entities/team'
require 'stbaldricks/entities/lib/fundraising_page'
require 'stbaldricks/entities/lib/venue'
require 'stbaldricks/entities/lib/permissions'
require 'stbaldricks/entities/lib/default_cacheable'
require 'stbaldricks/entities/concerns/donation_recipient_concern'
require 'stbaldricks/entities/concerns/event_year_concern'
require 'stbaldricks/entities/concerns/fundraising_page_concern'
require 'stbaldricks/entities/concerns/venue_concern'

module SBF
  module Client
    class Fundraiser < SBF::Client::TopLevelEntity
      include VenueConcern, FundraisingPageConcern, EventYearConcern, DonationRecipientConcern, Entities::DefaultCacheable
      endpoint SBF::Client::FundraiserEndpoint
      actions DEFAULT_CRUD_ACTIONS
      action :join_team
      action :leave_team
      action :join_event
      action :leave_event

      disallow_instantiation

      module Status
        ACTIVE = 'active'.freeze
        PENDING = 'pending'.freeze
      end

      module Category
        ATHLETICS = 'athletics'.freeze
        CELEBRATION = 'celebration'.freeze
        SALE = 'sale'.freeze
        OTHER = 'other'.freeze
      end

      module HowCreated
        WEBSITE = 'website'.freeze
        ADMIN_CONTROL_PANEL = 'admin_control_panel'.freeze
        ADMIN_IMPORT = 'admin_import'.freeze
        LEGACY = 'legacy'.freeze
        MOBILE = 'mobile'.freeze
      end

      module FundRelationshipType
        NO_ASSOCIATION = 'no_association'.freeze
        INDIRECT_ONE = 'indirect_one'.freeze
        INDIRECT_MULTIPLE = 'indirect_multiple'.freeze
        DIRECT = 'direct'.freeze
        DIRECT_BUT_LOCKED = 'direct_but_locked'.freeze
        INHERITED_FROM_TEAM = 'inherited_from_team'.freeze
        INHERITED_FROM_EVENT = 'inherited_from_event'.freeze
        INHERITED_FROM_TEAMS_EVENT = 'inherited_from_teams_event'.freeze
        INHERITED_FROM_BOTH_TEAM_AND_EVENT = 'inherited_from_both_team_and_event'.freeze
      end

      class Photos < SBF::Client::BaseEntity
        attr_reader :avatar, :default, :icon
      end

      class Policies < SBF::Client::BaseEntity
        attr_accessor :is_waiver_policy_accepted
      end

      class Totals < SBF::Client::BaseEntity
        attr_reader :amount_raised
        attr_reader :donations
      end

      class Rankings < SBF::Client::BaseEntity
        class Ranking < SBF::Client::BaseEntity
          attr_reader :rank, :total
        end

        entity_attr_reader :team, 'SBF::Client::Fundraiser::Rankings::Ranking'
        entity_attr_reader :event, 'SBF::Client::Fundraiser::Rankings::Ranking'
      end

      class FullEvent < SBF::Client::FullEvent; end

      class PartialEvent < SBF::Client::PartialEvent
        def to_hash
          entity_hash = super

          # Fundraisers do not require an event but are still reliant on an event year
          entity_hash.delete(:id) if id.nil?

          entity_hash
        end
      end

      def dirty_data(with_keys = false)
        data = super(with_keys)
        data[:event].delete(:id) if data[:event] && data[:event][:id].nil?
        data
      end
    end

    class PartialFundraiser < Fundraiser
      attr_accessor :id
    end

    class FullFundraiser < Fundraiser
      extend Forwardable
      def_delegator :@photos, :avatar, :avatar_photo_url

      attr_accessor :id

      attr_reader :created_at, :modified_at
      attr_reader :display_name

      attr_accessor :category
      attr_accessor :display_name_override
      attr_accessor :facebook_fundraiser_id
      attr_accessor :fund_relationship_type
      attr_accessor :fundraising_goal, :fundraising_message
      attr_accessor :how_created
      attr_accessor :modified_by
      attr_accessor :offline_donation_uuid, :offline_donation_form
      attr_accessor :start_date, :end_date
      attr_accessor :start_time, :end_time
      attr_accessor :status
      attr_accessor :title

      entity_attr_reader :permissions, 'SBF::Client::Permissions', nil, true
      entity_attr_reader :rankings, 'SBF::Client::Fundraiser::Rankings', nil, true
      entity_attr_reader :totals, 'SBF::Client::Fundraiser::Totals'

      entity_attr_accessor :event, 'SBF::Client::Fundraiser::FullEvent', 'SBF::Client::Fundraiser::PartialEvent', true
      entity_attr_accessor :fund, 'SBF::Client::FullFund', 'SBF::Client::PartialFund', true
      entity_attr_accessor :fundraising_page, 'SBF::Client::FundraisingPage', nil, true
      entity_attr_accessor :owner, 'SBF::Client::FullPerson', 'SBF::Client::PartialPerson'
      entity_attr_accessor :photos, 'SBF::Client::Fundraiser::Photos'
      entity_attr_accessor :policies, 'SBF::Client::Fundraiser::Policies'
      entity_attr_accessor :team, 'SBF::Client::FullTeam', 'SBF::Client::PartialTeam', true
      entity_attr_accessor :third_party_media, 'SBF::Client::ThirdPartyMedia', nil, true
      entity_attr_accessor :venue, 'SBF::Client::Venue'

      entity_collection_attr_accessor :milestones, 'SBF::Client::Milestone', nil, true

      def active?
        status == SBF::Client::Fundraiser::Status::ACTIVE
      end

      def past?
        return Date.today > Date.parse(end_date) if end_date

        Date.today > Date.parse(start_date) if start_date
      rescue
        nil
      end

      def event?
        event && !event.id.nil?
      end
    end
  end
end
