require 'stbaldricks/endpoints/participant'
require 'stbaldricks/entities/fund'
require 'stbaldricks/entities/event'
require 'stbaldricks/entities/team'
require 'stbaldricks/entities/person'
require 'stbaldricks/entities/lib/address'
require 'stbaldricks/entities/lib/base'
require 'stbaldricks/entities/lib/top_level'
require 'stbaldricks/entities/lib/fundraising_page'
require 'stbaldricks/entities/lib/name'
require 'stbaldricks/entities/lib/phone'
require 'stbaldricks/entities/lib/permissions'
require 'stbaldricks/entities/lib/default_cacheable'
require 'stbaldricks/entities/concerns/donation_recipient_concern'
require 'stbaldricks/entities/concerns/event_year_concern'
require 'stbaldricks/entities/concerns/fundraising_page_concern'

module SBF
  module Client
    class Participant < SBF::Client::TopLevelEntity
      include FundraisingPageConcern, EventYearConcern, DonationRecipientConcern, Entities::DefaultCacheable
      endpoint SBF::Client::ParticipantEndpoint
      actions DEFAULT_CRUD_ACTIONS
      action :join_team
      action :start_team
      action :leave_team
      action :delete_role
      action :move_to_different_event

      disallow_instantiation

      module Status
        ACTIVE = 'active'.freeze
        PENDING = 'pending'.freeze
      end

      module ShirtSize
        FAST_REG_UNSPECIFIED = 'fast_reg_unspecified'.freeze
        NOT_SPECIFIED = 'not specified'.freeze
        SAVE_FOR_RESEARCH = 'save_for_research'.freeze
        YOUTH_SMALL = 'youth_small'.freeze
        YOUTH_MEDIUM = 'youth_medium'.freeze
        YOUTH_LARGE = 'youth_large'.freeze
        XS = 'xs'.freeze
        SMALL = 'small'.freeze
        MEDIUM = 'medium'.freeze
        LARGE = 'large'.freeze
        XL = 'xl'.freeze
        XXL = 'xxl'.freeze
        XXXL = '3x'.freeze
        XXXXL = '4x'.freeze
        XXXXXL = '5x'.freeze

        def self.list
          %i(
            FAST_REG_UNSPECIFIED
            SAVE_FOR_RESEARCH
            YOUTH_SMALL
            YOUTH_MEDIUM
            YOUTH_LARGE
            XS
            SMALL
            MEDIUM
            LARGE
            XL
            XXL
            XXXL
            XXXXL
            XXXXXL
          )
        end
      end

      module HowCreated
        WEBSITE = 'website'.freeze
        ADMIN_IMPORT = 'admin_import'.freeze
        ADMIN_CONTROL_PANEL = 'admin_control_panel'.freeze
        MOBILE = 'mobile'.freeze
        FAST_REGISTRATION = 'fast_registration'.freeze
      end

      module FundRelationshipType
        NO_ASSOCIATION = 'no_association'.freeze
        DIRECT = 'direct'.freeze
        DIRECT_BUT_LOCKED = 'direct_but_locked'.freeze
        INHERITED_FROM_TEAM = 'inherited_from_team'.freeze
        INHERITED_FROM_EVENT = 'inherited_from_event'.freeze
        INHERITED_FROM_TEAMS_EVENT = 'inherited_from_teams_event'.freeze
        INHERITED_FROM_BOTH_TEAM_AND_EVENT = 'inherited_from_both_team_and_event'.freeze
      end

      class Photos < SBF::Client::BaseEntity
        attr_reader :before, :after, :avatar, :default, :icon
      end

      class Policies < SBF::Client::BaseEntity
        attr_accessor :is_waiver_policy_accepted, :is_razor_policy_accepted
      end

      class ShaveSchedule < SBF::Client::BaseEntity
        attr_accessor :date, :time, :notified
      end

      class Roles < SBF::Client::BaseEntity
        module Status
          ACTIVE = 'active'.freeze
          PENDING = 'pending'.freeze
          SUSPENDED = 'suspended'.freeze
        end

        module Type
          SHAVEE = 'shavee'.freeze
          BARBER = 'barber'.freeze
          TREASURER = 'treasurer'.freeze
          SHAVEE_RECRUITER = 'shavee_recruiter'.freeze
          ORGANIZER = 'organizer'.freeze
          TEAM_CAPTAIN = 'team_captain'.freeze
          LEAD_TREASURER = 'lead_treasurer'.freeze
          LEAD_ORGANIZER = 'lead_organizer'.freeze
          BARBER_COORDINATOR = 'barber_coordinator'.freeze
          EVENT_SUPPORT_VOLUNTEER = 'event_support_volunteer'.freeze
          INDIVIDUAL_FUNDRAISER = 'individual_fundraiser'.freeze
        end

        CORE_COMMITTEE_TYPES = [Type::ORGANIZER, Type::TREASURER, Type::BARBER_COORDINATOR, Type::SHAVEE_RECRUITER].freeze
        TYPES_WITHOUT_LEADS = [Type::SHAVEE, Type::EVENT_SUPPORT_VOLUNTEER, Type::INDIVIDUAL_FUNDRAISER, Type::SHAVEE_RECRUITER, Type::TEAM_CAPTAIN,
                               Type::BARBER_COORDINATOR, Type::BARBER, Type::TREASURER, Type::ORGANIZER].freeze
        class Role < SBF::Client::BaseEntity
          attr_accessor :status

          def active?
            status == SBF::Client::Participant::Roles::Status::ACTIVE
          end
        end

        class LeadRole < SBF::Client::Participant::Roles::Role
          attr_accessor :is_lead, :status
        end

        class Organizer < SBF::Client::Participant::Roles::LeadRole
          class Policies < SBF::Client::BaseEntity
            attr_accessor :agreement_received_date, :coaches_questionnaire_received_date
            attr_accessor :is_revenue_waiver_accepted, :is_host_agreement_waiver_accepted
          end

          class PreviousOrganization < SBF::Client::BaseEntity
            attr_accessor :charity, :date
          end

          attr_accessor :is_lead, :status
          entity_attr_accessor :policies, 'SBF::Client::Participant::Roles::Organizer::Policies'
          entity_attr_accessor :previous_organization, 'SBF::Client::Participant::Roles::Organizer::PreviousOrganization'
        end

        class Barber < SBF::Client::Participant::Roles::Role
          class License < SBF::Client::BaseEntity
            attr_accessor :number, :organization
          end

          class Salon < SBF::Client::BaseEntity
            attr_accessor :name
            entity_attr_accessor :phone, 'SBF::Client::Phone'
          end

          entity_attr_accessor :license, 'SBF::Client::Participant::Roles::Barber::License'
          entity_attr_accessor :salon, 'SBF::Client::Participant::Roles::Barber::Salon'
        end

        class Shavee < SBF::Client::Participant::Roles::Role
          attr_accessor :requested_shavee_packet
        end

        class EventSupportVolunteer < SBF::Client::Participant::Roles::Role; end
        class IndividualFundraiser < SBF::Client::Participant::Roles::Role; end
        class ShaveeRecruiter < SBF::Client::Participant::Roles::Role; end
        class TeamCaptain < SBF::Client::Participant::Roles::Role; end
        class BarberCoordinator < SBF::Client::Participant::Roles::Role; end
        class Treasurer < SBF::Client::Participant::Roles::LeadRole; end

        entity_attr_accessor :shavee, 'SBF::Client::Participant::Roles::Shavee'
        entity_attr_accessor :event_support_volunteer, 'SBF::Client::Participant::Roles::EventSupportVolunteer'
        entity_attr_accessor :individual_fundraiser, 'SBF::Client::Participant::Roles::IndividualFundraiser'
        entity_attr_accessor :barber_coordinator, 'SBF::Client::Participant::Roles::BarberCoordinator'
        entity_attr_accessor :shavee_recruiter, 'SBF::Client::Participant::Roles::ShaveeRecruiter'
        entity_attr_accessor :team_captain, 'SBF::Client::Participant::Roles::TeamCaptain'
        entity_attr_accessor :organizer, 'SBF::Client::Participant::Roles::Organizer'
        entity_attr_accessor :treasurer, 'SBF::Client::Participant::Roles::Treasurer'
        entity_attr_accessor :barber, 'SBF::Client::Participant::Roles::Barber'
      end

      class Totals < SBF::Client::BaseEntity
        attr_reader :amount_raised
      end

      class Rankings < SBF::Client::BaseEntity
        class Ranking < SBF::Client::BaseEntity
          attr_reader :rank, :total
        end

        entity_attr_reader :team, 'SBF::Client::Participant::Rankings::Ranking'
        entity_attr_reader :event, 'SBF::Client::Participant::Rankings::Ranking'
      end
    end

    class PartialParticipant < Participant
      attr_accessor :id
    end

    class FullParticipant < Participant
      extend Forwardable
      def_delegator :@photos, :avatar, :avatar_photo_url

      attr_accessor :id

      attr_reader :created_at, :modified_at
      attr_reader :display_name

      attr_accessor :display_name_override
      attr_accessor :facebook_fundraiser_id
      attr_accessor :fund_relationship_type
      attr_accessor :fundraising_goal, :fundraising_message
      attr_accessor :how_created
      attr_accessor :is_virtual
      attr_accessor :modified_by
      attr_accessor :offline_donation_uuid, :offline_donation_form
      attr_accessor :shirt_size
      attr_accessor :status

      entity_attr_reader :permissions, 'SBF::Client::Permissions', nil, true
      entity_attr_reader :permissions_at_event, 'SBF::Client::PermissionsAtEvent', nil, true
      entity_attr_reader :photos, 'SBF::Client::Participant::Photos'
      entity_attr_reader :rankings, 'SBF::Client::Participant::Rankings', nil, true
      entity_attr_reader :totals, 'SBF::Client::Participant::Totals'

      entity_attr_accessor :event, 'SBF::Client::FullEvent', 'SBF::Client::PartialEvent', true
      entity_attr_accessor :fund, 'SBF::Client::FullFund', 'SBF::Client::PartialFund', true
      entity_attr_accessor :fundraising_page, 'SBF::Client::FundraisingPage', nil, true
      entity_attr_accessor :person, 'SBF::Client::FullPerson'
      entity_attr_accessor :policies, 'SBF::Client::Participant::Policies'
      entity_attr_accessor :roles, 'SBF::Client::Participant::Roles', nil, true
      entity_attr_accessor :shave_schedule, 'SBF::Client::Participant::ShaveSchedule'
      entity_attr_accessor :team, 'SBF::Client::FullTeam', 'SBF::Client::PartialTeam', true
      entity_attr_accessor :third_party_media, 'SBF::Client::ThirdPartyMedia', nil, true

      entity_collection_attr_accessor :milestones, 'SBF::Client::Milestone', nil, true

      def active?
        status == SBF::Client::Participant::Status::ACTIVE
      end
    end
  end
end
