require 'stbaldricks/entities/lib/profile_address'
require 'stbaldricks/entities/lib/base'
require 'stbaldricks/entities/lib/top_level'
require 'stbaldricks/entities/lib/profile_email_address'
require 'stbaldricks/entities/lib/name'
require 'stbaldricks/entities/lib/phone'
require 'stbaldricks/enums/league_status'
require 'stbaldricks/entities/lib/opt_out_settings'
require 'stbaldricks/entities/lib/alternate_shipping_address'
require 'stbaldricks/entities/lib/default_cacheable'
require 'stbaldricks/entities/concerns/type_concern'

module SBF
  module Client
    class Person < SBF::Client::TopLevelEntity
      include TypeConcern, Entities::DefaultCacheable
      disallow_instantiation

      actions DEFAULT_CRUD_ACTIONS
      blacklist_action :delete

      module Status
        PENDING = 'pending'.freeze # 1
        ACTIVE = 'active'.freeze # 2
        DECEASED = 'deceased'.freeze # 7 (may be deprecated: working on this with SBF)
      end

      module Type
        INDIVIDUAL = 'individual'.freeze
        BUSINESS = 'business'.freeze
        FOUNDATION = 'foundation'.freeze

        IDS = {
          INDIVIDUAL => 1,
          BUSINESS => 2,
          FOUNDATION => 3
        }.freeze
        IDS.freeze

        def self.to_s(type_id)
          IDS.key(type_id)
        end

        def self.value_of(type)
          IDS[type]
        end
      end

      module Gender
        NOT_SELECTED = 'not_selected'.freeze
        CHOSE_NOT_TO_SPECIFY = 'chose_not_to_specify'.freeze
        MALE = 'male'.freeze
        FEMALE = 'female'.freeze
      end

      module HowCreated
        WEBSITE = 'website'.freeze
        ADMIN_IMPORT = 'admin_import'.freeze
        ADMIN_CONTROL_PANEL = 'admin_control_panel'.freeze
        SUBMIT_CASH = 'submit_cash'.freeze
        MOBILE = 'mobile'.freeze
        PARTICIPANT_REGISTRATION = 'participant_registration'.freeze
        FUNDRAISER_REGISTRATION = 'fundraiser_registration'.freeze
        FAST_REGISTRATION = 'fast_registration'.freeze
        FAST_DONATION = 'fast_donation'.freeze
        RECURRING_GIFTS = 'recurring'.freeze
        GIVE_ONCE = 'give_once'.freeze
        DONATE = 'donate'.freeze
        PAYMENT_SPRING_APP = 'payment_spring_app'.freeze
      end

      module ShippingPreference
        USE_PRIMARY_ADDRESS = 'use_primary_address'.freeze
        USE_VENUE_ADDRESS = 'use_venue_address'.freeze
        USE_ALTERNATE_ADDRESS = 'use_alternate_address'.freeze
      end

      class Addresses < SBF::Client::BaseEntity
        entity_attr_accessor :primary, 'SBF::Client::ProfileAddress'
        entity_attr_accessor :secondary, 'SBF::Client::ProfileAddress'
      end

      class PhoneNumbers < SBF::Client::BaseEntity
        entity_attr_accessor :primary, 'SBF::Client::Phone'
        entity_attr_accessor :secondary, 'SBF::Client::Phone'
      end

      class EmailAddresses < SBF::Client::BaseEntity
        entity_attr_accessor :primary, 'SBF::Client::ProfileEmailAddress'
        entity_attr_accessor :secondary, 'SBF::Client::ProfileEmailAddress'
      end

      class Policies < SBF::Client::BaseEntity
        BACKGROUND_CHECK_YEARS_VALID = 5
        attr_accessor :is_photo_policy_accepted
        attr_accessor :background_check_date, :background_check_url

        def background_check_valid_at_date?(check_date)
          check_date = DateTime.parse(check_date) if check_date.is_a?(String)
          check_date = DateTime.parse(check_date.to_s) if check_date.is_a?(Date)
          check_date = DateTime.parse(check_date.strftime('%d/%m/%Y %H:%M:%S')) if check_date.is_a?(Time)
          background_check_date && check_date.is_a?(DateTime) \
                                && background_check_date > check_date.prev_year(SBF::Client::Person::Policies::BACKGROUND_CHECK_YEARS_VALID)
        end
      end

      class HowAffected < SBF::Client::BaseEntity
        attr_accessor :cancer_type, :family_relationship, :cancer_status
        attr_reader :is_child_survivor, :is_child_family, :is_child_friend, :is_adult_survivor, :is_adult_friend_or_family, :is_none

        module CancerType
          ACUTE_LYMPHOBLASTIC_LEUKEMIA = 'Acute lymphoblastic leukemia (ALL)'.freeze
          ACUTE_MYELOID_LEUKEMIA = 'Acute myeloid leukemia (AML)'.freeze
          BRAIN_OR_SPINAL_CORD_TUMOR = 'Brain or spinal cord tumor'.freeze
          CLEAR_CELL_SARCOMA_OF_THE_KIDNEY = 'Clear cell sarcoma of the kidney'.freeze
          EWING_SARCOMA = 'Ewing sarcoma'.freeze
          GERM_CELL_TUMOR = 'Germ cell tumor'.freeze
          HEPATOBLASTOMA = 'Hepatoblastoma'.freeze
          HODGKIN_LYMPHOMA = 'Hodgkin lymphoma'.freeze
          LANGERHANS_CELL_HISTIOCYTOSIS = 'Langerhans Cell Histiocytosis (LCH)'.freeze
          LYMPHOMA = 'Lymphoma'.freeze
          NEUROBLASTOMA = 'Neuroblastoma'.freeze
          NON_HODGKIN_LYMPHOMA = 'Non-Hodgkin lymphoma'.freeze
          OSTEOSARCOMA = 'Osteosarcoma'.freeze
          RETINOBLASTOMA = 'Retinoblastoma'.freeze
          RHABDOMYOSARCOMA = 'Rhabdomyosarcoma'.freeze
          WILMS_OR_OTHER_KIDNEY_TUMOR = 'Wilms or other kidney tumor'.freeze
          ADULT_WITH_A_TYPE_OF_CHILDHOOD_CANCER = 'Adult with a type of childhood cancer'.freeze
          OTHER_CHILDHOOD_CANCER = 'Other childhood cancer'.freeze
          @list = {
            ACUTE_LYMPHOBLASTIC_LEUKEMIA => 'acute_lymphoblastic_leukemia',
            ACUTE_MYELOID_LEUKEMIA => 'acute_myeloid_leukemia',
            BRAIN_OR_SPINAL_CORD_TUMOR => 'brain_or_spinal_cord_tumor',
            CLEAR_CELL_SARCOMA_OF_THE_KIDNEY => 'kidney_sarcoma',
            EWING_SARCOMA => 'ewing_sarcoma',
            GERM_CELL_TUMOR => 'germ_cell_tumor',
            HEPATOBLASTOMA => 'hepatoblastoma',
            HODGKIN_LYMPHOMA => 'hodgkin_lymphoma',
            LANGERHANS_CELL_HISTIOCYTOSIS => 'langerhans_cell_histiocytosis',
            LYMPHOMA => 'lymphoma',
            NEUROBLASTOMA => 'neuroblastoma',
            NON_HODGKIN_LYMPHOMA => 'non_hodgkin_lymphoma',
            OSTEOSARCOMA => 'osteosarcoma',
            RETINOBLASTOMA => 'retinoblastoma',
            RHABDOMYOSARCOMA => 'rhabdomyosarcoma',
            WILMS_OR_OTHER_KIDNEY_TUMOR => 'wilms_or_other_kidney_tumor',
            ADULT_WITH_A_TYPE_OF_CHILDHOOD_CANCER => 'adult_with_childhood_cancer',
            OTHER_CHILDHOOD_CANCER => 'other_childhood_cancer'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        module FamilyRelationship
          NONE = 'None'.freeze
          MOTHER = 'Mother'.freeze
          FATHER = 'Father'.freeze
          SISTER = 'Sister'.freeze
          BROTHER = 'Brother'.freeze
          AUNT = 'Aunt'.freeze
          UNCLE = 'Uncle'.freeze
          GRANDMOTHER = 'Grandmother'.freeze
          GRANDFATHER = 'Grandfather'.freeze
          STEP_MOTHER = 'Step-Mother'.freeze
          STEP_FATHER = 'Step-Father'.freeze
          LEGAL_GUARDIAN = 'Legal-Guardian'.freeze
          @list = {
            NONE => 'none',
            MOTHER => 'mother',
            FATHER => 'father',
            SISTER => 'sister',
            BROTHER => 'brother',
            AUNT => 'aunt',
            UNCLE => 'uncle',
            GRANDMOTHER => 'grandmother',
            GRANDFATHER => 'grandfather',
            STEP_MOTHER => 'stepmother',
            STEP_FATHER => 'stepfather',
            LEGAL_GUARDIAN => 'legal_guardian'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        module CancerStatus
          CANCER_FREE = 'Cancer free'.freeze
          IN_REMISSION = 'In remission'.freeze
          IN_TREATMENT = 'In treatment'.freeze
          IN_MAINTENANCE = 'In maintenance'.freeze
          LOST_TO_CANCER = 'Lost to cancer'.freeze
          @list = {
            CANCER_FREE => 'cancer_free',
            IN_REMISSION => 'in_remission',
            IN_TREATMENT => 'in_treatment',
            IN_MAINTENANCE => 'in_maintenance',
            LOST_TO_CANCER => 'lost_to_cancer'
          }.freeze

          class << self
            attr_reader :list
          end
        end
      end
      class Occupation < SBF::Client::BaseEntity
        module Type
          STUDENT = 'Student'.freeze
          EDUCATOR = 'Educator'.freeze
          PHYSICIAN = 'Physician'.freeze
          NURSE = 'Nurse'.freeze
          RESEARCHER = 'Researcher'.freeze
          SOCIAL_WORKER = 'Social Worker'.freeze
          OTHER_CARE_GIVER = 'Other Caregiver'.freeze
          CHILDHOOD_HEALTH_ADVOCATE = 'Childhood Health Advocate'.freeze
          FIREFIGHTER = 'Firefighter'.freeze
          POLICE_OFFICER = 'Police Officer'.freeze
          MILITARY = 'Military'.freeze
          INSURANCE_INDUSTRIES = 'Insurance Industries'.freeze
          BANKING_FINANCE = 'Banking & Finance'.freeze
          SECURITIES_TRADING = 'Securities & Trading'.freeze
          BARBER_COSMETOLOGIST = 'Barber or Cosmetologist'.freeze
          OTHER = 'Other'.freeze

          @list = {
            STUDENT => 'student',
            EDUCATOR => 'educator',
            PHYSICIAN => 'physician',
            NURSE => 'nurse',
            RESEARCHER => 'researcher',
            SOCIAL_WORKER => 'social_worker',
            OTHER_CARE_GIVER => 'other_care_giver',
            CHILDHOOD_HEALTH_ADVOCATE => 'childhood_health_advocate',
            FIREFIGHTER => 'firefighter',
            POLICE_OFFICER => 'police_officer',
            MILITARY => 'military',
            INSURANCE_INDUSTRIES => 'insurance_industries',
            BANKING_FINANCE => 'banking_finance',
            SECURITIES_TRADING => 'securities_trading',
            BARBER_COSMETOLOGIST => 'barber_cosmetologist',
            OTHER => 'other'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        module EducatorType
          PRESCHOOL = 'Preschool'.freeze
          ELEMENTARY_SCHOOL = 'Elementary School'.freeze
          MIDDLE_SCHOOL = 'Middle School'.freeze
          HIGH_SCHOOL = 'High School'.freeze
          COLLEGE = 'College'.freeze

          @list = {
            PRESCHOOL => 'preschool',
            ELEMENTARY_SCHOOL => 'elementary_school',
            MIDDLE_SCHOOL => 'middle_school',
            HIGH_SCHOOL => 'high_school',
            COLLEGE => 'college'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        module StudentType
          ELEMENTARY_SCHOOL = 'Elementary School'.freeze
          MIDDLE_SCHOOL = 'Middle School'.freeze
          HIGH_SCHOOL = 'High School'.freeze
          COLLEGE = 'College'.freeze
          @list = {
            ELEMENTARY_SCHOOL => 'elementary_school',
            MIDDLE_SCHOOL => 'middle_school',
            HIGH_SCHOOL => 'high_school',
            COLLEGE => 'college'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        module MilitaryBranchType
          ARMY = 'Army'.freeze
          NAVY = 'Navy'.freeze
          AIR_FORCE = 'Air Force'.freeze
          MARINES = 'Marines'.freeze
          COAST_GUARD = 'Coast Guard'.freeze
          NATIONAL_GUARD = 'National Guard'.freeze

          @list = {
            ARMY => 'army',
            NAVY => 'navy',
            AIR_FORCE => 'air_force',
            MARINES => 'marines',
            COAST_GUARD => 'coast_guard',
            NATIONAL_GUARD => 'national_guard'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        module MilitaryStatusType
          ACTIVE_DUTY = 'Active Duty'.freeze
          RESERVES = 'Reserves'.freeze
          RETIRED = 'Retired'.freeze

          @list = {
            ACTIVE_DUTY => 'active_duty',
            RESERVES => 'reserves',
            RETIRED => 'retired'
          }.freeze

          class << self
            attr_reader :list
          end
        end

        class EducationDetails < SBF::Client::BaseEntity
          attr_accessor :level, :school_name, :is_fraternity_sorority, :fraternity_sorority_name
        end

        class MilitaryDetails < SBF::Client::BaseEntity
          attr_accessor :branch, :status, :unit
        end

        class FinanceDetails < SBF::Client::BaseEntity
          attr_accessor :industry, :business_name, :job_title
        end

        attr_accessor :type
        attr_reader :is_student, :is_educator, :is_military, :is_finance, :is_other

        # TODO: Need to rework this. The details hash needs to contain enough information to be
        # able to determine what type it should be. As it is currently written, it cannot instanitate itself
        # without the information contained in the SBF::Client::Person::Occupation instance.
        attr_accessor :details
        def details=(value)
          details_will_change!
          if educator? || student?
            @details = SBF::Client::Person::Occupation::EducationDetails.new(value)
          elsif military?
            @details = SBF::Client::Person::Occupation::MilitaryDetails.new(value)
          elsif finance? || other?
            @details = SBF::Client::Person::Occupation::FinanceDetails.new(value)
          end
        end
      end

      class Totals < SBF::Client::BaseEntity
        attr_reader :amount_raised
        attr_reader :donations
      end

      def league_status_title
        case league_status
        when SBF::Client::LeagueStatus::SQUIRE
          'Squire of Hope'
        when SBF::Client::LeagueStatus::KNIGHT
          'Knight of the Bald Table'
        when SBF::Client::LeagueStatus::KNIGHT_COMMANDER
          'Knight Commander'
        when SBF::Client::LeagueStatus::CRUSADER
          'Crusader for a Cure'
        when SBF::Client::LeagueStatus::BARON
          'Baron of the Brave'
        when SBF::Client::LeagueStatus::NOT_APPLICABLE
          ''
        end
      end
    end

    class PartialPerson < SBF::Client::Person
      attr_accessor :id, :type
    end

    class FullPerson < SBF::Client::Person
      attr_accessor :id
      attr_accessor :salesforce_id
      attr_accessor :salesforce_account_id
      attr_accessor :type
      attr_accessor :status
      attr_accessor :gender
      attr_accessor :is_minor
      attr_accessor :birthday
      attr_reader :display_name, :first_and_last_name
      attr_accessor :display_name_override
      attr_accessor :years_as_participant, :years_as_organizer
      attr_reader :league_status
      attr_reader :name
      entity_attr_accessor :name_pieces, 'SBF::Client::NamePieces'
      entity_attr_accessor :addresses, 'SBF::Client::Person::Addresses'
      entity_attr_accessor :phone_numbers, 'SBF::Client::Person::PhoneNumbers'
      entity_attr_accessor :email_addresses, 'SBF::Client::Person::EmailAddresses'
      entity_attr_accessor :occupation, 'SBF::Client::Person::Occupation', nil, true
      entity_attr_accessor :policies, 'SBF::Client::Person::Policies'
      entity_attr_accessor :how_affected, 'SBF::Client::Person::HowAffected', nil, true
      entity_attr_accessor :opt_out_settings, 'SBF::Client::OptOutSettings', nil, true
      attr_accessor :selected_shipping_method
      entity_attr_accessor :alternate_shipping_address, 'SBF::Client::AlternateShippingAddress', nil, true
      attr_reader :created_at, :modified_at
      attr_accessor :modified_by
      attr_accessor :how_created
      entity_attr_reader :totals, 'SBF::Client::Person::Totals', nil, true

      def active?
        status == SBF::Client::Person::Status::ACTIVE
      end

      # Dynamically build convenience methods for checking the person's league status (i.e. 'squire?', 'knight?', etc.)
      SBF::Client::LeagueStatus.to_a.delete_if { |x| x == SBF::Client::LeagueStatus::NOT_APPLICABLE }.each do |league_status|
        define_method "#{league_status}?".to_sym do
          self.league_status == league_status
        end
      end
    end
  end
end
