require 'stbaldricks/endpoints/recurring_gift'
require 'stbaldricks/entities/fund'
require 'stbaldricks/entities/event'
require 'stbaldricks/entities/team'
require 'stbaldricks/entities/participant'
require 'stbaldricks/entities/fundraiser'
require 'stbaldricks/entities/person'
require 'stbaldricks/entities/organization'
require 'stbaldricks/entities/lib/address'
require 'stbaldricks/entities/lib/top_level'
require 'stbaldricks/entities/lib/name'
require 'stbaldricks/entities/lib/payment'
require 'stbaldricks/entities/lib/default_cacheable'
require 'stbaldricks/entities/concerns/model_type_concern'

module SBF
  module Client
    class RecurringGift < SBF::Client::TopLevelEntity
      include Entities::DefaultCacheable
      endpoint SBF::Client::RecurringGiftEndpoint
      actions DEFAULT_CRUD_ACTIONS
      blacklist_action :aggregate
      blacklist_action :delete
      action :sync
      action :change_billing_day
      action :change_amount

      attr_accessor :id

      module Status
        UNKNOWN = 'unknown'.freeze
        PAST_DUE = 'past_due'.freeze
        ACTIVE = 'active'.freeze
        CANCELED = 'canceled'.freeze
        PENDING = 'pending'.freeze
      end

      module RecipientType
        FUND = 'fund'.freeze
        EVENT = 'event'.freeze
        TEAM = 'team'.freeze
        PARTICIPANT = 'participant'.freeze
        FUNDRAISER = 'fundraiser'.freeze
        MEMORIAL = 'memorial'.freeze
      end

      module DonorType
        INDIVIDUAL = 'individual'.freeze
        BUSINESS = 'business'.freeze
        ORGANIZATION_FOUNDATION = 'foundation'.freeze
      end

      module FundingType
        APPLEPAY = 'applepay'.freeze
        CREDIT_CARD = 'credit_card'.freeze
        GOOGLE_PAY = 'google_pay'.freeze
        PAYPAL = 'paypal'.freeze
      end

      class FullParticipant < SBF::Client::FullParticipant
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::PARTICIPANT
        end
      end

      class PartialParticipant < SBF::Client::PartialParticipant
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::PARTICIPANT
        end
      end

      class FullFundraiser < SBF::Client::FullFundraiser
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::FUNDRAISER
        end
      end

      class PartialFundraiser < SBF::Client::PartialFundraiser
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::FUNDRAISER
        end
      end

      class FullTeam < SBF::Client::FullTeam
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::TEAM
        end
      end

      class PartialTeam < SBF::Client::PartialTeam
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::TEAM
        end
      end

      class FullEvent < SBF::Client::FullEvent
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::EVENT
        end
      end

      class PartialEvent < SBF::Client::PartialEvent
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::EVENT
        end
      end

      class FullFund < SBF::Client::FullFund
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::FUND
        end
      end

      class PartialFund < SBF::Client::PartialFund
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::FUND
        end
      end

      class FullMemorial < SBF::Client::FullMemorial
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::MEMORIAL
        end
      end

      class PartialMemorial < SBF::Client::PartialMemorial
        include ModelTypeConcern
        attr_reader :model_type

        def model_type
          SBF::Client::RecurringGift::RecipientType::MEMORIAL
        end
      end

      class Tribute < SBF::Client::BaseEntity
        attr_accessor :type
        attr_reader :name
        entity_attr_accessor :name_pieces, 'SBF::Client::NamePieces'

        module Type
          HONOR = 'honor'.freeze
          MEMORY = 'memory'.freeze
        end
      end
    end

    class PartialRecurringGift < SBF::Client::RecurringGift
      attr_accessor :id
    end
    class FullRecurringGift < SBF::Client::RecurringGift
      attr_accessor :status
      attr_accessor :event_year
      attr_accessor :amount
      attr_accessor :fee_amount
      attr_accessor :billing_day_of_month
      attr_accessor :last_collected_at
      attr_accessor :last_failed_at
      attr_accessor :display_name
      attr_accessor :display_name_override
      attr_accessor :is_unrecognized
      attr_accessor :promotional_code
      attr_accessor :funding_type
      attr_accessor :payment_gateway
      attr_accessor :customer_id
      attr_accessor :subscription_id
      entity_collection_attr_accessor :donations, 'SBF::Client::FullDonation', 'SBF::Client::PartialDonation'
      attr_accessor :created_at, :modified_at, :modified_by, :how_created, :how_created_id, :past_due_at
      entity_attr_accessor :tribute, 'SBF::Client::RecurringGift::Tribute'

      multitype_attr_accessor(
        :payment_details,
        [
          [
            ->(v) { v[:type] == SBF::Client::Payment::Type::APPLEPAY },
            'SBF::Client::Payment::ApplepayDetails'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Payment::Type::CREDIT_CARD },
            'SBF::Client::Payment::SubscriptionCreditCardDetails'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Payment::Type::GOOGLEPAY },
            'SBF::Client::Payment::GooglePayDetails'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Payment::Type::NONCE },
            'SBF::Client::Payment::NonceDetails'
          ],
          [
            ->(v) { v[:type] == SBF::Client::Payment::Type::PAYPAL },
            'SBF::Client::Payment::PaypalDetails'
          ]
        ]
      )

      multitype_attr_accessor(
        :recipient,
        [
          [
            ->(v) { v[:model_type] == SBF::Client::RecurringGift::RecipientType::PARTICIPANT },
            'SBF::Client::RecurringGift::FullParticipant',
            'SBF::Client::RecurringGift::PartialParticipant'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::RecurringGift::RecipientType::FUNDRAISER },
            'SBF::Client::RecurringGift::FullFundraiser',
            'SBF::Client::RecurringGift::PartialFundraiser'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::RecurringGift::RecipientType::TEAM },
            'SBF::Client::RecurringGift::FullTeam',
            'SBF::Client::RecurringGift::PartialTeam'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::RecurringGift::RecipientType::EVENT },
            'SBF::Client::RecurringGift::FullEvent',
            'SBF::Client::RecurringGift::PartialEvent'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::RecurringGift::RecipientType::FUND },
            'SBF::Client::RecurringGift::FullFund',
            'SBF::Client::RecurringGift::PartialFund'
          ],
          [
            ->(v) { v[:model_type] == SBF::Client::RecurringGift::RecipientType::MEMORIAL },
            'SBF::Client::RecurringGift::FullMemorial',
            'SBF::Client::RecurringGift::PartialMemorial'
          ]
        ],
        optional: true
      )

      multitype_attr_accessor(
        :donor,
        [
          [
            ->(v) { v[:type] == SBF::Client::RecurringGift::DonorType::INDIVIDUAL },
            'SBF::Client::FullPerson',
            'SBF::Client::PartialPerson'
          ],
          [
            ->(v) { v[:type] == SBF::Client::RecurringGift::DonorType::BUSINESS },
            'SBF::Client::FullOrganization',
            'SBF::Client::PartialOrganization'
          ],
          [
            ->(v) { v[:type] == SBF::Client::RecurringGift::DonorType::ORGANIZATION_FOUNDATION },
            'SBF::Client::FullOrganization',
            'SBF::Client::PartialOrganization'
          ]
        ],
        optional: true
      )

      def amount=(value)
        amount_will_change!
        @amount = value.to_s.gsub(/[^0-9\.]/, '').to_f
      end

      def fee_amount=(value)
        fee_amount_will_change!
        @fee_amount = value.to_s.gsub(/[^0-9\.]/, '').to_f
      end
    end
  end
end
