require 'httparty'
require 'singleton'
require 'stbaldricks/api'
require 'stbaldricks/configuration'
require 'stbaldricks/log'
require 'uri'

# Require all monkey patches
patch_dir = File.realpath("#{File.dirname(__FILE__)}/patches/")
Dir["#{patch_dir}/*.rb"].sort.each { |file| require file }

# Require all entities
entity_dir = File.realpath("#{File.dirname(__FILE__)}/entities/")
Dir["#{entity_dir}/**/*.rb"].sort.each { |file| require file }

# Require all enums
enum_dir = File.realpath("#{File.dirname(__FILE__)}/enums/")
Dir["#{enum_dir}/*.rb"].sort.each { |file| require file }

# Client library class for accessing the St. Baldrick's Foundation API
module SBF
  module Client
    module Api
      module Request
        include HTTParty

        base_uri SBF::Client::Configuration.base_uri

        # Makes a HTTP GET request to the specified path with specified params and authentication credentials based on the configured client class.
        # This method isn't intended to be used by a user and is for internal use only
        #
        # @param path [string]
        # @param params [Hash]
        # @return [HTTParty::Response]
        def self.get_request(path, params = {})
          raise SBF::Client::InvalidURIError, "Invalid URI: #{path}" unless valid_uri?(path)

          request_log('GET', path, params) do
            get(path, options(query: params, auth: user_token || general_token))
          end
        end

        # Makes a HTTP POST request to the specified path with specified params and authentication credentials based on the configured client class.
        # Raises an error if the client library isn't configured with user specific credentials.
        # This method isn't intended to be used by a user and is for internal use only
        #
        # @param path [string]
        # @param params [Hash]
        # @return [HTTParty::Response]
        # @raise [Error]
        def self.post_request(path, params = {})
          raise SBF::Client::InvalidURIError, "Invalid URI: #{path}" unless valid_uri?(path)

          request_log('POST', path, params) do
            post(path, options(body: params.to_json, type: :post, auth: user_token || general_token))
          end
        end

        def self.file_post_request(path:, params: {}, file:, filename:)
          raise SBF::Client::InvalidURIError, "Invalid URI: #{path}" unless valid_uri?(path)
          raise SBF::Client::Error, 'Not yet authenticated' if user_token.nil?
          raise SBF::Client::Error, 'Invalid File' unless file.is_a?(File) || file.is_a?(Tempfile)

          request_log('POST', path, params) do
            post(path, options(body: params.merge(file: file, original_filename: filename), type: :file_post, auth: user_token))
          end
        end

        def self.options(auth: nil, type: :get, query: nil, body: nil)
          {}.tap do |options|
            options.merge!(
              base_uri: SBF::Client::Configuration.base_uri,
              headers: {}.tap do |hsh|
                hsh['Accept'] = 'application/json'
                hsh['X-Request-Id'] = SBF::Client::Configuration.request_id
                hsh['X-Forwarded-For'] = SBF::Client::Configuration.forwarded_for unless SBF::Client::Configuration.forwarded_for.blank?
              end,
              verify: verify_ssl_peer?
            )
            options[:body] = body if body
            options[:query] = query if query
            options[:headers]['Content-Type'] = 'application/json' if type == :post
            options[:headers]['Content-Type'] = 'multipart/form-data' if type == :file_post
            options[:headers]['Authorization'] = "Bearer #{auth}" if auth
          end
        end

        def self.general_token
          SBF::Client::Configuration.general_token
        end
        private_class_method :general_token

        def self.user_token
          SBF::Client::Configuration.user_token
        end
        private_class_method :user_token

        def self.request_log(type, path, params)
          SBF::Client::LOG.debug { "SBF::Client - #{type}: #{path} | Body: #{JSON.pretty_generate(params)}" }
          response = yield
          SBF::Client::LOG.debug { "SBF::Client - Response: #{response}" }
          response
        end
        private_class_method :request_log

        def self.valid_uri?(path)
          URI.parse(path)
          true
        rescue URI::InvalidURIError
          false
        end
        private_class_method :valid_uri?

        def self.verify_ssl_peer?
          (ENV['VERIFY_SSL_PEER'] || true).to_b
        end
        private_class_method :verify_ssl_peer?
      end
    end
  end
end
